/*++

Module Name:

    ioctl.c


X36 IO handler.

Overview:

  This module is responsible for handling all the device control calls that
pass through this filter.  It performs the following functions:
  -> Tweaking the USB descriptors as they come up through the stack to add
     the parameters necessary to create a virtual keyboard.
  -> Capturing an I/O request to reply with keyboard data on my own timers
  -> Replying with keyboard data
  -> Filter X36 joystick data

--*/

//#define DBG 1

#include <WDM.H>
#include "usbioctl.h"
#include "usbdi.h"
#include "usbdlib.h"
#include "filter.h"
#include "fakehid.h"




#ifdef ALLOC_PRAGMA
        #pragma alloc_text(PAGE, VA_Ioctl)
#endif


NTSTATUS VA_Ioctl(struct DEVICE_EXTENSION *  devExt, PIRP Irp)
{
    NTSTATUS            ntStatus = STATUS_SUCCESS;    
    PIO_STACK_LOCATION  IrpStack = IoGetCurrentIrpStackLocation(Irp);    
    ULONG dwControlCode = IrpStack->Parameters.DeviceIoControl.IoControlCode;
  PURB pUrb = (PURB)IrpStack->Parameters.Others.Argument1;

          
    if (IOCTL_INTERNAL_USB_SUBMIT_URB == dwControlCode)
  {

    switch (pUrb->UrbHeader.Function)
    {
        case URB_FUNCTION_BULK_OR_INTERRUPT_TRANSFER:
        if ((devExt->ReadIrp == NULL)&&(KeGetCurrentIrql()==DISPATCH_LEVEL)) // Don't capture if we're not really running.
        {
//          DbgPrn(("Capturing keyboard read IRP.\n"));
          
          KeAcquireSpinLockAtDpcLevel(&devExt->ReadLock);
          devExt->ReadIrp = Irp;
          KeReleaseSpinLockFromDpcLevel(&devExt->ReadLock);
          
                IncrementPendingActionCount(devExt);
                
                IoMarkIrpPending(Irp);
                ntStatus = STATUS_PENDING;
                
                if (devExt->NewData)
                {
                  DbgPrn(("Filling request right away.  New data: %d\n",devExt->NewData));
                  KeInsertQueueDpc(&devExt->ReturnDPC,NULL,NULL);
                }
              }
              else
              {
                IoCopyCurrentIrpStackLocationToNext(Irp);
            IoSetCompletionRoutine(Irp, &ReportCompletion, NULL, TRUE, TRUE, TRUE);    
            ntStatus = IoCallDriver(devExt->topDevObj, Irp);
          }
        break;
    
        case URB_FUNCTION_RESET_PIPE:
          DbgPrn(("Reset pipe.\n"));
        if ( devExt->ReadIrp != NULL )
        {
          KeCancelTimer(&devExt->Timer);
          CompleteRead( NULL, devExt, NULL, NULL);        
        }
        break ;
        
      case URB_FUNCTION_ABORT_PIPE:
        DbgPrn(("Abort pipe.\n"));
        KeCancelTimer(&devExt->Timer);
        KeAcquireSpinLockAtDpcLevel(&devExt->ReadLock);
        if (devExt->ReadIrp != NULL)        
        {
            devExt->ReadIrp->IoStatus.Status = STATUS_CANCELLED;
            IoCompleteRequest(devExt->ReadIrp, IO_NO_INCREMENT);  
          devExt->ReadIrp = NULL;
          DecrementPendingActionCount( devExt );
        }
        KeReleaseSpinLockFromDpcLevel(&devExt->ReadLock);       
        break;
    
        case URB_FUNCTION_GET_DESCRIPTOR_FROM_DEVICE:
        IoCopyCurrentIrpStackLocationToNext(Irp) ;
        IoSetCompletionRoutine(Irp,(PIO_COMPLETION_ROUTINE)ConfigDescriptorCompletion, NULL, TRUE, TRUE,TRUE) ;
        ntStatus = IoCallDriver(devExt->topDevObj, Irp) ;
        break;
    
         case URB_FUNCTION_GET_DESCRIPTOR_FROM_INTERFACE:
        IoCopyCurrentIrpStackLocationToNext(Irp) ;
        IoSetCompletionRoutine(Irp,(PIO_COMPLETION_ROUTINE)ReportDescriptorCompletion, NULL, TRUE,TRUE, TRUE) ;
        ntStatus = IoCallDriver(devExt->topDevObj, Irp) ;
        break;
        
      default:
        IoSkipCurrentIrpStackLocation(Irp);
        ntStatus = IoCallDriver(devExt->topDevObj, Irp);
      }
      
  }
  else
  {
    IoSkipCurrentIrpStackLocation(Irp);
    ntStatus = IoCallDriver(devExt->topDevObj, Irp);
  }

    return ntStatus;    
}


NTSTATUS ConfigDescriptorCompletion( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context )
{
  /* Should print the data in the system. */
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    struct DEVICE_EXTENSION *  DeviceExtension = DeviceObject->DeviceExtension;
    PURB pUrb = (PURB)IoGetCurrentIrpStackLocation(Irp)->Parameters.Others.Argument1;
    struct _URB_CONTROL_TRANSFER   *pControlTransfer = (struct _URB_CONTROL_TRANSFER *) pUrb;
    PUCHAR BufPtr=NULL;
    PUCHAR Temp;
    unsigned long b,i,pos, tmp;
    
    
    if (Irp->PendingReturned) {
        IoMarkIrpPending( Irp );
    }
    

  // OK,we need to (a) snapture the report id; then (b) snapture the interrupts themselves.
  // To get the report ID we need to know which one to watch for.
  // So, first we have to capture the HID descriptor and parse that.  It comes between the configuration and interface descriptors.


  if (pUrb != NULL)
  {
      // Figure out what packet this was.
      
    DbgPrn(("  SetupPacket          :"));
    for(b=0; b<sizeof(pControlTransfer->SetupPacket); b++)
      DbgPrn((" %02x", pControlTransfer->SetupPacket[b]));
    DbgPrn(("\n"));

    BufPtr=NULL;
    if ((PUCHAR)pControlTransfer->TransferBuffer)
    {
      BufPtr=(PUCHAR)pControlTransfer->TransferBuffer;
    }
    else if(pControlTransfer->TransferBufferMDL)
    {
      BufPtr = (PUCHAR)MmGetSystemAddressForMdl(pControlTransfer->TransferBufferMDL);
    }
    if (BufPtr==NULL)
    { // Confused.  Abort.
      DbgPrn(("Can't find buffer, aborting.\n"));
      return STATUS_SUCCESS;
    }
                
    if ((pControlTransfer->SetupPacket[3]==0x02)&&((pControlTransfer->SetupPacket[6]>0x09)||(pControlTransfer->SetupPacket[7]>0)))
    {
      DbgPrn(("Complete descriptor transferred.  Locating HID descriptor.\n"));
      
      for(pos=0;((BufPtr[pos+1]!=0x21)&&(pos<pControlTransfer->TransferBufferLength));pos=pos+BufPtr[pos]);
      
      if (pos>=pControlTransfer->TransferBufferLength)
      {
        DbgPrn(("Can't find HID descriptor, aborting.\n"));
        return STATUS_SUCCESS;
      }
      else
      {
        DbgPrn(("HID descriptor located at position %x.\n",pos));
      }
        
      DeviceExtension->ReportDescType = BufPtr[pos+6];
      DeviceExtension->ReportBufSize = BufPtr[pos+7];
      BufPtr[pos+7]+=sizeof(MyReportDescriptor)+2;
                    
                    
      
      if (Temp = (PUCHAR)MmGetSystemAddressForMdl(pControlTransfer->TransferBufferMDL))
        RtlCopyMemory(Temp,BufPtr,pControlTransfer->TransferBufferLength);
    }   
  }
  return STATUS_SUCCESS;
}



NTSTATUS ReportDescriptorCompletion( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context )
{
  /* Should print the data in the system. */
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    struct DEVICE_EXTENSION *  DeviceExtension = DeviceObject->DeviceExtension;
    PURB pUrb = (PURB)IoGetCurrentIrpStackLocation(Irp)->Parameters.Others.Argument1;
    struct _URB_CONTROL_TRANSFER   *pControlTransfer = (struct _URB_CONTROL_TRANSFER *) pUrb;
    PUCHAR BufPtr=NULL;
    PUCHAR Temp;
    unsigned long b,i,pos, tmp;
    
    
    if (Irp->PendingReturned) {
        IoMarkIrpPending( Irp );
    }
    

  // OK,we need to (a) snapture the report id; then (b) snapture the interrupts themselves.
  // To get the report ID we need to know which one to watch for.
  // So, first we have to capture the HID descriptor and parse that.  It comes between the configuration and interface descriptors.


  if (pUrb != NULL)
  {
      // Figure out what packet this was.
      
    DbgPrn(("  SetupPacket          :"));
    for(b=0; b<sizeof(pControlTransfer->SetupPacket); b++)
      DbgPrn((" %02x", pControlTransfer->SetupPacket[b]));
    DbgPrn(("\n"));

    BufPtr=NULL;
    if ((PUCHAR)pControlTransfer->TransferBuffer)
    {
      BufPtr=(PUCHAR)pControlTransfer->TransferBuffer;
    }
    else if(pControlTransfer->TransferBufferMDL)
    {
      BufPtr = (PUCHAR)MmGetSystemAddressForMdl(pControlTransfer->TransferBufferMDL);
    }
    if (BufPtr==NULL)
    { // Confused.  Abort.
      DbgPrn(("Can't find buffer, aborting.\n"));
      return STATUS_SUCCESS;
    }
                


    if (pControlTransfer->SetupPacket[3]==DeviceExtension->ReportDescType)
    {
      DbgPrn(("Report Descriptor.\n"));
    
      // Fix throttle type.
      BufPtr[35]=0x32;
      
      // HID PARSER SHOULD GO HERE
      // in the mean time, just kill the autocalibration.     
      
      // x min
      *((short*)(&BufPtr[9])) = 0;
      // x max
      *((short*)(&BufPtr[12])) = 0x0FFF;
      
      // y min
      *((short*)(&BufPtr[23])) = 0;
      // y max
      *((short*)(&BufPtr[26])) = 0x0FFF;

      // throttle min
      BufPtr[37] = 0;
      // throttle max
      *((short*)(&BufPtr[39])) = 0xFF;

      // rudder min
      BufPtr[50] = 0;
      // rudder max
      *((short*)(&BufPtr[52])) = 0xFF;

      // rot1 min
      BufPtr[63] = 0;
      // rot1  max
      *((short*)(&BufPtr[65])) = 0xFF;
      
      // rot2 min
      BufPtr[76] = 0;
      // rot2 max
      *((short*)(&BufPtr[78])) = 0xFF;


      DbgPrn(("Report descriptor data normalized.\n"));

        
      
      // move rear portion of decriptor over
      RtlMoveMemory(BufPtr+8,BufPtr+6,pControlTransfer->TransferBufferLength-6);
      // insert the 'I am report 1' bytes
      BufPtr[6]=0x85;     BufPtr[7]=0x01; 
      pControlTransfer->TransferBufferLength+=2;
      // now add the extra report descripts
      RtlCopyMemory(&BufPtr[pControlTransfer->TransferBufferLength],MyReportDescriptor, sizeof(MyReportDescriptor));
      pControlTransfer->TransferBufferLength+=sizeof(MyReportDescriptor);
      
      // finally, turn the mouse ticker on
      KeSetTimer( &DeviceExtension->MouseTimer, RtlConvertLongToLargeInteger(-1* MOUSE_TICK), &DeviceExtension->MouseDPC );
    }
  }
  return STATUS_SUCCESS;
}



NTSTATUS ReportCompletion( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context )
{
  /* Should print the data in the system. */
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    struct DEVICE_EXTENSION *  DeviceExtension = DeviceObject->DeviceExtension;
    PURB pUrb = (PURB)IoGetCurrentIrpStackLocation(Irp)->Parameters.Others.Argument1;
  struct _URB_BULK_OR_INTERRUPT_TRANSFER *pBulkOrInterruptTransfer = (struct _URB_BULK_OR_INTERRUPT_TRANSFER *) pUrb;
    PUCHAR BufPtr=NULL;
    PUCHAR Temp;
    LARGE_INTEGER LocalTime;
    QENTRY Current;
    unsigned char Delta[0x0a];
    unsigned char switch1,switch2,shift;
    unsigned char i;
    unsigned long b,pos, tmp;
    unsigned short axisval,outaxisval;

    
    if (Irp->PendingReturned) {
        IoMarkIrpPending( Irp );
    }

    
    // Lag avoidance routine.
  // KeQuerySystemTime(&LocalTime);
  

  // OK,we need to (a) snapture the report id; then (b) snapture the interrupts themselves.
  // To get the report ID we need to know which one to watch for.
  // So, first we have to capture the HID descriptor and parse that.  It comes between the configuration and interface descriptors.
  // DbgPrn(("Joystick sending data packet.\n"));

  if ((pUrb != NULL)&&(pBulkOrInterruptTransfer->TransferBufferLength == 0x0b))
  {
    // This is the stick reporting its packet.
    // DbgPrn(("Stick reporting packet.\n"));
    
    BufPtr = NULL;
  
    if ((PUCHAR)pBulkOrInterruptTransfer->TransferBuffer)
    {
      BufPtr=(PUCHAR)pBulkOrInterruptTransfer->TransferBuffer;
    }
    else if(pBulkOrInterruptTransfer->TransferBufferMDL)
    {
      BufPtr = (PUCHAR)MmGetSystemAddressForMdl(pBulkOrInterruptTransfer->TransferBufferMDL);
    }
    if (BufPtr==NULL)
    { // Confused.  Abort.
      return STATUS_SUCCESS;
    }
    
    
    // Compute data structures.
    // Store system time.
    KeQuerySystemTime(&LocalTime);

                        
                
    // Process axes.
    // x and y are special cases; do them first
    i=0;
    axisval = (unsigned short)BufPtr[0] + (((unsigned short)BufPtr[1]&0x0F) << 8);
    outaxisval = Calculate(&(DeviceExtension->Calibration[i]),axisval,12);
    BufPtr[0] = (unsigned char) (outaxisval & 0x00FF);
    BufPtr[1] = (BufPtr[1]&0xF0) + (unsigned char)((outaxisval>>8) & 0x0F);
//    DbgPrint("X axis: %d %d %d %d %d\n",axisval,outaxisval,DeviceExtension->Calibration[i].value[0],DeviceExtension->Calibration[i].value[1],DeviceExtension->Calibration[i].value[2]);

    //y
    i++;
    axisval = (((unsigned short)BufPtr[1]&0xF0) >> 4) + ((unsigned short)BufPtr[2] << 4);
    outaxisval = Calculate(&(DeviceExtension->Calibration[i]),axisval,12);
    BufPtr[1] = (BufPtr[1]&0x0F) + (unsigned char)(outaxisval & 0x000F << 4);
    BufPtr[2] = (unsigned char) ((outaxisval & 0x0FF0) >> 4);
//    DbgPrint("Y axis: %d %d %d %d %d\n",axisval,outaxisval,DeviceExtension->Calibration[i].value[0],DeviceExtension->Calibration[i].value[1],DeviceExtension->Calibration[i].value[2]);



    // everything else
    for (i++;i<NUM_AXIS; i++)
    {
      axisval = (unsigned short) BufPtr[i+1];
      outaxisval = Calculate(&(DeviceExtension->Calibration[i]),axisval,8);
      BufPtr[i+1] = (unsigned char) outaxisval;

//      DbgPrint("Axis %d: %d %d %d %d %d\n",i,axisval,outaxisval,DeviceExtension->Calibration[i].value[0],DeviceExtension->Calibration[i].value[1],DeviceExtension->Calibration[i].value[2]);     
    }
      
      
    DbgPrn(("Axis processing complete.\n"));                        
                

        // Right-shift hat switches 2 bits, for ease of computation.
        b=*(unsigned long*)&BufPtr[7];
        b=(b&0xF0003FFF) | ((b&0x03FFC000)<<2);
        
    // Convert hat switches to 8-way.  Compute hat switch deltas.
    for(i=0;i<3;i++)
    {
        shift = (unsigned char) ((b >> ((i*4)+16))&0x0F);
        shift = hat4to8(shift);
        pos=(0xFFFFFFFF ^ (0x0F << ((i*4)+16)));
        b = (b&pos)+ (shift <<((i*4)+16));
      }
    *(unsigned long*)&BufPtr[7] = b;


    // Compute deltas.
    for(i=0;i<0x09;i++)
    {
      Delta[i]=DeviceExtension->LastPosition[i] ^ BufPtr[i];
      DeviceExtension->LastPosition[i] = BufPtr[i];
    }
    for(;i<0x0b;i++)
    {
      Delta[i]=DeviceExtension->LastPosition[i];
      DeviceExtension->LastPosition[i] = BufPtr[i];
    }
        
    // Data structures computed.  Begin engine.
/*
    pos=*(unsigned long*)&BufPtr[7];
    DbgPrint("Buttons: %X ",pos);
    pos=*(unsigned long*)&Delta[7];
    DbgPrint("Delta: %X\n",pos);
*/    

    KeAcquireSpinLockAtDpcLevel(&DeviceExtension->ScriptLock);


    // Script queueing engine.
    // Buttons
    for(i=0;i<8;i++)
    {
      if ((Delta[7+(i/8)]>>(i%8))&0x01) // Button toggled
        Queue(DeviceExtension,i,DeviceExtension->Buttons[i][((BufPtr[7+(i/8)]>>(i%8))&0x01)],LocalTime);                
    }
    
    // 3-ways - special handling to enforce ordered operation
    // release must be handled before press, so queue presses before releases - queue is lifo
    for(i=8;i<14;i++)
    {
      if (((Delta[7+(i/8)]>>(i%8))&0x01)&&(((BufPtr[7+(i/8)]>>(i%8))&0x01)==1)) // Button toggled
        Queue(DeviceExtension,i,DeviceExtension->Buttons[i][((BufPtr[7+(i/8)]>>(i%8))&0x01)],LocalTime);                
    }
    for(i=8;i<14;i++)
    {
      if (((Delta[7+(i/8)]>>(i%8))&0x01)&&(((BufPtr[7+(i/8)]>>(i%8))&0x01)==0)) // Button toggled
        Queue(DeviceExtension,i,DeviceExtension->Buttons[i][((BufPtr[7+(i/8)]>>(i%8))&0x01)],LocalTime);                
    }

    


    //DbgPrn(("Begin hat switch processing.\n"));
    
    b=*(unsigned long*)&BufPtr[7];
    tmp=*(unsigned long*)&Delta[7];
    for(i=0;i<4;i++) // Hat switches.
    {
      unsigned char value, prev;
                        
      // Grab hat values
      value = (unsigned char) ((b >> ((i*4)+16))&0x0F);
      prev = (unsigned char) ((tmp >> ((i*4)+16))&0x0F);
      //DbgPrint("Hat %d:  b %x   prev %x.\n",i,value,prev);
      
      if (prev != value )
      {
        // Execute release before press, so queue press first.
        Queue(DeviceExtension,(unsigned char)(14+(9*i)+value),DeviceExtension->Buttons[14+(9*i)+value][1],LocalTime);
        Queue(DeviceExtension,(unsigned char)(14+(9*i)+prev),DeviceExtension->Buttons[14+(9*i)+prev][0],LocalTime);
        
      }

    }         


    KeReleaseSpinLockFromDpcLevel(&DeviceExtension->ScriptLock);

        // Move packet forward one byte and mod to indicate joystick vs keyboard
    RtlMoveMemory(&BufPtr[1],BufPtr,0x0b/*pBulkOrInterruptTransfer->TransferBufferLength*/);
    BufPtr[0]=0x01;   
    pBulkOrInterruptTransfer->TransferBufferLength/*+=1*/ = 0x0c;

    // Patch on internal DirectX buffer.
    RtlCopyMemory(&BufPtr[8],&DeviceExtension->KeyboardStatus[21],4);
    
    
    //
    // Respawn timer if we have any events.
    //
    QueuePeek(&DeviceExtension->ActionQueue, &Current);
    if (Current.Head)
    {
          KeSetTimer( &DeviceExtension->Timer, Current.Time, &DeviceExtension->TimerDPC );              
    }
      
  }
  //DbgPrn(("JoyPendingActionCount: %d\n",DeviceExtension->pendingActionCount));
  return STATUS_SUCCESS;
}


void CompleteRead( IN PKDPC Dpc, IN PVOID DeferredContext, IN PVOID SystemArgument1, IN PVOID SystemArgument2)
{
  struct DEVICE_EXTENSION * DeviceExtension = (struct DEVICE_EXTENSION *)DeferredContext;
  PIRP Irp = NULL;
    PIO_STACK_LOCATION IrpStack;
    ULONG bytesRead;
    UCHAR buffer[30];
    unsigned char switch1,switch2,shift,b;
    unsigned long i,pos;
    

  if (DeviceExtension)
  {
    KIRQL OldIrqL;
        
    KeAcquireSpinLockAtDpcLevel(&DeviceExtension->ReadLock);
    Irp = DeviceExtension->ReadIrp;
    DeviceExtension->ReadIrp = NULL;
    KeReleaseSpinLockFromDpcLevel(&DeviceExtension->ReadLock);
    
    if (Dpc == &DeviceExtension->MouseDPC)
      KeSetTimer( &DeviceExtension->MouseTimer, RtlConvertLongToLargeInteger(-1 * MOUSE_TICK), &DeviceExtension->MouseDPC );        
  }

  if (Irp)
  {
    IrpStack = IoGetCurrentIrpStackLocation(Irp);

    if ((Dpc == &DeviceExtension->MouseDPC) && *(short*)&DeviceExtension->MouseStatus[1] && *(short*)&DeviceExtension->MouseStatus[3])
    {
      // Mouse movement packet
//      DbgPrn(("Completing mouse movement read..\n"));
//      DbgPrn(("Mouse data:  Buttons: %X  X: %d  Y:  %d\n",DeviceExtension->MouseStatus[0],*(short*)&DeviceExtension->MouseStatus[1],*(short*)&DeviceExtension->MouseStatus[3]));
      buffer[0]=0x03;
      RtlCopyMemory(&buffer[1],DeviceExtension->MouseStatus,sizeof(DeviceExtension->MouseStatus));
      bytesRead=sizeof(DeviceExtension->MouseStatus)+1;
      DeviceExtension->NewData&= ~(0x04);
    }
    else if (DeviceExtension->NewData & 0x04) 
    {
//      DbgPrn(("Completing mouse read..\n"));
      DbgPrn(("Mouse data:  Buttons: %X  X: %d  Y:  %d\n",DeviceExtension->MouseStatus[0],*(short*)&DeviceExtension->MouseStatus[1],*(short*)&DeviceExtension->MouseStatus[3]));
      buffer[0]=0x03;
      RtlCopyMemory(&buffer[1],DeviceExtension->MouseStatus,1);
      RtlFillMemory(&buffer[2],sizeof(DeviceExtension->MouseStatus)-1,0x00);
      bytesRead=sizeof(DeviceExtension->MouseStatus)+1;
//      if ((*(short*)&DeviceExtension->MouseStatus[1] == 0) && (*(short*)&DeviceExtension->MouseStatus[3]==0))
      DeviceExtension->NewData&= ~(0x04);
    }   
    else if (DeviceExtension->NewData & 0x02)
    { 
//      DbgPrn(("Completing keyboard read..\n"));
        
      RtlCopyMemory(&buffer[1],DeviceExtension->KeyboardStatus,29);
      buffer[0]=0x02;
    
      RtlFillMemory(&buffer[20],4,0x00);  
  
      // Complete our IO report.
          bytesRead = 30;     
          DeviceExtension->NewData&=0xFD;
    }
    else
    {     
//      DbgPrn(("Completing joystick read..\n"));
      
      buffer[0]=0x01;
      RtlCopyMemory(&buffer[1],DeviceExtension->LastPosition,0x08);
      RtlCopyMemory(&buffer[8],&DeviceExtension->KeyboardStatus[21],4);     
      bytesRead=0x0c;
      DeviceExtension->NewData&=0xFE;
    }
          
    if (bytesRead > IrpStack->Parameters.DeviceIoControl.OutputBufferLength)
      {
          bytesRead = IrpStack->Parameters.DeviceIoControl.OutputBufferLength;
      }
  
    RtlCopyMemory((PUCHAR) Irp->UserBuffer, buffer, bytesRead);
      Irp->IoStatus.Information = bytesRead;
      Irp->IoStatus.Status = STATUS_SUCCESS;
    
      IoCompleteRequest(Irp, IO_NO_INCREMENT);
      DecrementPendingActionCount(DeviceExtension);
  }
}

// Stub function for Iotimer event to send periodic keyboard reports regardless of stick status.
void KeyboardTimer(IN PDEVICE_OBJECT DeviceObject, IN PVOID Context)
{
    CompleteRead( NULL, Context, NULL, NULL); 
}



/* Utilitiy functions to convert hat direction.
   Compiled inline to save performance. */
unsigned char hat4to8(unsigned char four)
{
  switch(four)
  {
    case 0: return 0;
    case 1: return 1;
    case 2: return 3;
    case 3: return 2;
    case 4: return 5;
    case 6: return 4;
    case 8: return 7;
    case 9: return 8;
    case 12: return 6;
  }
  return 0;
}


unsigned char hat8to4(unsigned char eight)
{
  switch(eight)
  {
    case 0: return 0;
    case 1: return 1;
    case 2: return 3;
    case 3: return 2;
    case 4: return 6;
    case 5: return 4;
    case 6: return 12;
    case 7: return 8;
    case 8: return 9;
  }
  return 0;
}
