// QUEUE.C
//
// Handles the timer queue for the X36 scripting engine.
// 
// includes:
//
// queueinit: inits queue.
// scriptadd: adds a whole script to the queue.
// pop: removes the first item from the queue.
// peek: returns the first item in the queue without destroying it
// queueclear: destroys a queue.

//#define DBG 1
#include <WDM.H>
#include "filter.h"


NTSTATUS QueueInit(PXQUEUE MyQueue)
{
	unsigned int i;
	
	if (MyQueue == NULL)
		return STATUS_INVALID_PARAMETER;
	
	MyQueue->Head = NULL;
	KeInitializeSpinLock(&MyQueue->Lock);
	MyQueue->Next = 0; MyQueue->Last = 0;
	for (i=0; i<256; i++)
		MyQueue->Free[i]=(unsigned char)i;
	
	
	return STATUS_SUCCESS;
}

NTSTATUS AddScript(PXQUEUE MyQueue, QENTRY NewEntry)
{
	PQENTRY QNew, QCurrent, QPrev;
	
	
	DbgPrn(("Entering AddScript.\n"));
	if (MyQueue == NULL)
		return STATUS_INVALID_PARAMETER;
	
	if (NewEntry.Current == NULL)
		return STATUS_SUCCESS;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);
	DbgPrn(("Spinlock aquired.\n"));

	if (MyQueue->Next - MyQueue->Last < 255)
	{
		DbgPrn(("Assigning new space in buffer.."));
		QNew = &MyQueue->Buffer[MyQueue->Free[MyQueue->Next]];
		MyQueue->Next++;
		DbgPrn(("..assigned.  Position: %d     Next: %d  NextPos: %d  Last: %d  LastPos: %d\n",MyQueue->Free[MyQueue->Next-1],MyQueue->Next,MyQueue->Free[MyQueue->Next],MyQueue->Last,MyQueue->Free[MyQueue->Last]));
		RtlCopyMemory(QNew,&NewEntry,sizeof(QENTRY));
		QNew->Next = NULL;
		DbgPrn(("Queuing item:  %x  Head: %x   Current: %x   Time: %x %x\n",QNew,QNew->Head,QNew->Current,QNew->Time.HighPart,QNew->Time.LowPart));		
		
		if (MyQueue->Head == NULL)
			MyQueue->Head = QNew;
		else
		{
			QCurrent=MyQueue->Head;
			QPrev=NULL;			
			
			while((QCurrent != NULL) && (RtlLargeIntegerGreaterThan(QNew->Time,QCurrent->Time)))
			{
				QPrev = QCurrent;
				QCurrent = QCurrent -> Next;
			}
			
			DbgPrn(("Queue position:   Head: %x  Prev: %x   New: %x   Current: %x\n",MyQueue->Head,QPrev,QNew,QCurrent));
			if (QPrev != NULL)
			{
				QNew->Next = QCurrent;
				QPrev->Next = QNew;
			}
			else
			{
				MyQueue->Head = QNew;
				QNew->Next = QCurrent;				
			}
			DbgPrn(("Queue status:   Head: %x   Next: %x\n",MyQueue->Head,(MyQueue->Head==NULL)?NULL:MyQueue->Head->Next));
		}		
	}	
	else
	{
		DbgPrn(("Addscript aborted due to insufficent resources.\n"));
		KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
		DbgPrn(("Spinlock released.\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
	DbgPrn(("Spinlock released.\n"));
	return STATUS_SUCCESS;
}



void QueuePop(PXQUEUE MyQueue, PQENTRY Output)
{
	PQENTRY QCurrent, QReturn;
	KIRQL OldIrqL;
	
	if ((MyQueue == NULL) || (Output==NULL))
		return;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);

	QCurrent = MyQueue->Head;
/*	if (QCurrent == QCurrent->Next)
		DbgBreakPoint();*/
	if (QCurrent != NULL)
	{
		RtlCopyMemory(Output,QCurrent,sizeof(QENTRY));	
		MyQueue->Head = QCurrent->Next;
		MyQueue->Free[MyQueue->Last]=(QCurrent - &MyQueue->Buffer[0]);//sizeof(QENTRY);
		DbgPrn(("Next: %d  NextPos: %d  Last: %d  LastPos: %d\n",MyQueue->Next,MyQueue->Free[MyQueue->Next],MyQueue->Last,MyQueue->Free[MyQueue->Last]));
		MyQueue->Last++;
	}
	else
		Output->Head=NULL;
	
	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
}

void QueuePeek(PXQUEUE MyQueue, PQENTRY Output)
{
	if ((MyQueue == NULL) || (Output==NULL))
		return;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);
	
	if (MyQueue->Head != NULL)
		RtlCopyMemory(Output,MyQueue->Head,sizeof(QENTRY));
	else
		Output->Head = NULL;
	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
}

void QueueClear(PXQUEUE MyQueue)
{
	PQENTRY QCurrent;
	KIRQL OldIrqL;	
	
	if (MyQueue == NULL)
		return;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);

	while(MyQueue->Head != NULL)
	{
		QCurrent = MyQueue->Head;
		MyQueue->Head = QCurrent->Next;
		MyQueue->Free[MyQueue->Last]=(QCurrent - &MyQueue->Buffer[0])/sizeof(QENTRY);		
		MyQueue->Last++;
	}
	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
}


void Queue(struct DEVICE_EXTENSION *devExt, unsigned char Frobbable, unsigned int Command, LARGE_INTEGER Time)
{
    QENTRY QNew;
    
    DbgPrn(("Queing command: %d\n",Command));
    Command = Command%MAX_COMMAND;
				
	if (devExt->Commands[Command] != NULL)
	{
		QNew.Head = QNew.Current = devExt->Commands[Command];
		QNew.Trigger = Frobbable;
		QNew.Time = RtlLargeIntegerAdd(RtlConvertLongToLargeInteger(QNew.Head->Delay),Time);
		AddScript(&devExt->ActionQueue, QNew);
	}
}